package houghtransform;
import ij.IJ;
import ij.ImagePlus;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;

import java.awt.Color;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

/**
 * Houghtransformation (HT) for circles. 
 * 
 * @author Roy
 *
 */
public class HoughCircles {

	//hough accumulator
	private int[][][] houghSpace;
	
	//nr of bins in alpha and d direction
	private int nrAlphaBins,nrDBins;
	
	//the image to perform
	private ImagePlus imp;
	
	
	//temporal variables of accumulateHoughCircle() for efficiency
	private double d,e;
	private int dBucket, dBucket2;
	
	private boolean stop;
	
	/**
	 * Constructor.
	 * Initialize HT for the image imp. The quantification of houghspace/ 
	 * count of vote buckets/bins is given by alphaBins, dBins.  
	 * @param imp Image to perform HT on, should be an binary image (at least 8-bit).
	 * @param alphaBins Number of bins in alpha direction.
	 * @param dBins Number of bins in d direction.
	 */
	public HoughCircles(ImagePlus imp, int radien){
		stop=false;
		nrDBins=imp.getHeight();
		nrAlphaBins=imp.getWidth();
		this.imp=imp;
		
		
		//create houghspace
		houghSpace=new int[nrAlphaBins][nrDBins][radien+1];
	}
	/**
	 * Returns a list of all non black pixel positions.
	 * @return List of point.
	 */
	public List<Point> getNonBlackPixels(){
		List<Point> l=new ArrayList<Point>();
		imp.updateImage();
		for(int y=0;y<imp.getHeight();y++)
		{

			for(int x=0;x<imp.getWidth();x++)
			{
				if(imp.getPixel(x,y)[0]!=0)//pixel non-black?
				{
					l.add(new Point(x,y));
				}
			}
		}
		return l;
	}
	
	/**
	 * Fills the accumulator array,
	 */
	public void fillHoughSpace(int radius,int radmin)
	{
		//for every non-black pixel accumulate a Hough circle FORMEL
		for(int y=0;y<imp.getHeight();y++)
		{
			if (y % 10 == 0)
				IJ.showProgress(y, imp.getHeight());
			if(stop){
				stop=false;
				return;
			}
			for(int x=0;x<imp.getWidth();x++)
			{
				if(imp.getPixel(x,y)[0]!=0)//pixel non-black?
				{
					accumulateHoughCircle(x,y,radius,radmin);
				}
			}
		}	
	}
	/**
	 * Stops filling the accumulator array, because computation
	 * can be expensive. Useful to stop threads running this algorithm.
	 */
	public void stop(){
		stop=true;
	}
	
	private void paint(double d,double e,int radius)
	{
		dBucket=(int)d;
		dBucket2=(int)e;
		if(dBucket>=0&&dBucket<nrAlphaBins && dBucket2>=0&&dBucket2<nrDBins) houghSpace[dBucket][dBucket2][radius]++;
	}
	/**
	 * Draws one hough circle for a specified image pixel x,y in accumulator array.
	 * @param x  X Coordinate of image pixel.
	 * @param y  Y Coordinate of image pixel.
	 */
	public void accumulateHoughCircle(int xCenter,int yCenter,int radius,int radmin)
	{
		int x, y, r2;
		r2 = radius * radius;
		d=xCenter;
		e=yCenter + radius;
		paint(d,e,radius-radmin);
		
		d=xCenter;
		e=yCenter - radius;
		paint(d,e,radius-radmin);
		
		d=xCenter + radius;
		e=yCenter ;
		paint(d,e,radius-radmin);
		
		d=xCenter - radius;
		e=yCenter ;
		paint(d,e,radius-radmin);
		
		y = radius;
		x = 1;
		y = (int) (Math.sqrt(r2 - 1) + 0.5);
		while (x < y) {
				d=xCenter + x;
				e=yCenter + y;
				paint(d,e,radius-radmin);
				d=xCenter + x;
				e=yCenter - y;
				paint(d,e,radius-radmin);
				d=xCenter - x;
				e=yCenter + y;
				paint(d,e,radius-radmin);
				d=xCenter - x;
				e=yCenter - y;
				paint(d,e,radius-radmin);
				d=xCenter + y;
				e=yCenter + x;
				paint(d,e,radius-radmin);;
				d=xCenter + y;
				e=yCenter - x;
				paint(d,e,radius-radmin);
				d=xCenter - y;
				e=yCenter + x;
				paint(d,e,radius-radmin);
				d=xCenter - y;
				e=yCenter - x;
				paint(d,e,radius-radmin);
			
			    x += 1;
			    y = (int) (Math.sqrt(r2 - x*x) + 0.5);
		}
		if (x == y) {
			d=xCenter + x;
			e=yCenter + y;
			paint(d,e,radius-radmin);
			d=xCenter + x;
			e=yCenter - y;
			paint(d,e,radius-radmin);
			d=xCenter - x;
			e=yCenter + y;
			paint(d,e,radius-radmin);
			d=xCenter - x;
			e= yCenter - y;
			paint(d,e,radius-radmin);
		}
	}
		
	
	/**
	 * Returns a visual correct result of the houghspace.
	 * @return Houghspace as ImagePlus.
	 */
	public ImagePlus getHoughSpaceImage(int place)
	{
		int[] fpixels = new int[nrAlphaBins * nrDBins];
		for (int a = 0; a < nrAlphaBins; a++) {
			for (int dBucket = 0; dBucket < nrDBins; dBucket++) {
				fpixels[dBucket * nrAlphaBins + a] = houghSpace[a][dBucket][place];
			}
		}
		FloatProcessor houghFP = new FloatProcessor(nrAlphaBins, nrDBins,fpixels);
		
		ImagePlus houghIP=new ImagePlus(imp.getTitle()+" Houghspace",houghFP); 
		return houghIP;
	}
	
	/**
	 * Find buckit/bin with maximal votes and return count.
	 * in abhngigkeit zum radius
	 * @return Count of maximal votes
	 */
	private double getNrMaximalVotes(int radmin,int radmax)
	{
		double maxVotes=0;
		double dump=0;
		for (int radius=radmin;radius<=radmax;radius++){
			for (int a = 0; a < nrAlphaBins; a++) {
				for (int dBucket = 0; dBucket < nrDBins; dBucket++) {
					dump=houghSpace[a][dBucket][radius-radmin];
					dump=dump*100/(radius*2*Math.PI);
					maxVotes=Math.max(maxVotes, dump);
				}
			}
		}
		
		return maxVotes;
	}
	
	/**
	 * Gets a list of maximum votes in houghspace by relative intensity,
	 * that mean every point in houghspace which votecount is greater
	 * relInt * nr. maximal votes.
	 */
	public HashMap<Integer,List<Point>>getMaximaByRelativeIntensity(double relInt,int radmin,int radmax)
	{
	
		relInt=Math.min(1.0,Math.max(relInt, 0.0));
		
		HashMap<Integer,List<Point>> maxima=new HashMap<Integer,List<Point>>();
		
		double maxVotes=getNrMaximalVotes(radmin, radmax);
		double voteThreshold=(double)(maxVotes*relInt);
		double dump;
		for (int radius=radmin;radius<=radmax;radius++){
		
			List<Point> maximalist=new ArrayList<Point>();
			for (int a = 0; a < nrAlphaBins; a++) {
				for (int d = 0; d < nrDBins; d++) {
					dump=houghSpace[a][d][radius-radmin];
					dump=dump*100/(radius*2*Math.PI);
					if(dump>voteThreshold){
						maximalist.add(new Point(a,d));
					}
				}
			}
			if (maximalist!=null){
				
				maxima.put(radius, maximalist);
				
			}
		}
		return maxima;
	}

		
	/**
	 * Returns an image with marked houspace maxima given by maxima.
	 * @param maxima List of maxima points.
	 * @return Image with maxima pixel.
	 */
	public ImagePlus getMaximaImage(HashMap<Integer, List<Point>> maxima, int radmin,int shownradius){
		
		int[] fpixels = new int[nrAlphaBins * nrDBins];			
		int a,dBucket;
		  List<Point> value =maxima.get(shownradius); 
		  Iterator<Point> it=value.iterator();
		while(it.hasNext()){				
			Point p=it.next();
			a=p.x;
			dBucket=p.y;
			fpixels[dBucket * nrAlphaBins + a] = houghSpace[a][dBucket][shownradius-radmin];		
		}
		
		FloatProcessor houghFP = new FloatProcessor(nrAlphaBins, nrDBins,fpixels);
		
		ImagePlus houghIP=new ImagePlus(imp.getTitle()+" Maxima",houghFP); 
		return houghIP;
	}
	
	
	/**
	 * Suppress maxima in local neighborhood.
	 * @param maxima List of maxima.
	 * @param maximaImage Image consisting of maxima pixel.
	 * @param maxThreshold Suppress only maxima which where maxThreshold smaller.
	 * @return List of rest of maxima.
	 */
	public HashMap<Integer,List<Point>>nonMaximumSupression(HashMap<Integer,List<Point>> maxima,ImagePlus maximaImage,double maxThreshold){
		
		
		HashMap<Integer,List<Point>> nonMaxSupMaxima=new HashMap<Integer,List<Point>>(maxima);
		
		
		FloatProcessor maxIm=((FloatProcessor)maximaImage.getProcessor());
		Iterator<Entry<Integer, List<Point>>> entries=nonMaxSupMaxima.entrySet().iterator();
		while(entries.hasNext()){
			Entry<Integer, List<Point>> thisEntry = (Entry<Integer, List<Point>>) entries.next();
			  List<Point> value = (List<Point>) thisEntry.getValue();
			  Iterator<Point> it=value.iterator();
		
			while(it.hasNext()){
				Point p=it.next();
				float v=maxIm.getPixelValue(p.x, p.y);
				if(	!(maxIm.getPixelValue(p.x+1, p.y)+maxThreshold<v&&
						maxIm.getPixelValue(p.x+1, p.y-1)+maxThreshold<v&&
						maxIm.getPixelValue(p.x+1, p.y+1)+maxThreshold<v&&
						maxIm.getPixelValue(p.x-1, p.y)+maxThreshold<v&&
						maxIm.getPixelValue(p.x-1, p.y-1)+maxThreshold<v&&
						maxIm.getPixelValue(p.x-1, p.y+1)+maxThreshold<v&&
						maxIm.getPixelValue(p.x, p.y+1)+maxThreshold<v&&
						maxIm.getPixelValue(p.x, p.y-1)+maxThreshold<v)){

					it.remove();
						}
			}	
		}
		return nonMaxSupMaxima;
	}
	
	
	/**
	 * Reduces the list of maximas to count elements.
	 * @param maxima List of maximas.
	 * @param maximaImage Image consisting of maxima pixel.
	 * @param count Count of maximas remaining.
	 * @return Reduced maxima list.
	 */
	public List<Point> reduceMaxima(List<Point> maxima,ImagePlus maximaImage,int count)
	{
		final FloatProcessor maxIm=((FloatProcessor)maximaImage.getProcessor());
		
		List<Point> reducedMaxima=new ArrayList<Point>(maxima);
		
		Collections.sort(reducedMaxima, new Comparator<Point>(){

			@Override
			public int compare(Point o1, Point o2) {
				
				float v1=maxIm.getPixelValue(o1.x, o1.y);
				float v2=maxIm.getPixelValue(o2.x, o2.y);
				return v1==v2 ? 0 : (v1<v2 ? -1 : +1);
			}
			
		});
		
		if(reducedMaxima.size()>count)
			reducedMaxima.subList(count, reducedMaxima.size()).clear();
		
		return reducedMaxima;
	}
	
	
	/**
	 * Builds an image with the input image in the background
	 * an the fitted red lines specified by the list of maximas. 
	 * @param maxima List of maxima.
	 * @return Image visualizing result of houghtransformation.
	 */
	public ImagePlus getImageWithHoughCircles(HashMap<Integer,List<Point>> maxima){
		
		ImageProcessor imgWHL = imp.getProcessor().duplicate().convertToRGB();
		Iterator<Entry<Integer, List<Point>>> entries=maxima.entrySet().iterator();
		while(entries.hasNext()){
			Entry<Integer, List<Point>> thisEntry = (Entry<Integer, List<Point>>) entries.next();
			  int key = (Integer) thisEntry.getKey();
			  List<Point> value = (List<Point>) thisEntry.getValue();
			  Iterator<Point> it=value.iterator();
			  while(it.hasNext()){
				Point p=it.next();
				imgWHL.setColor(Color.red);
				imgWHL.drawOval((int)p.x-key, (int)p.y-key, key*2, key*2);
			}
		}
		return new ImagePlus(imp.getTitle()+" Hough circles",imgWHL);
	}
}
